
--[[
    This file is a library to setup a lua 5.1 restricted environement
    ----

    Copyright (C) 2008-2011  TsT <tst@worldmaster.fr>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
]]--

--[[
  Hello,
  If you found a bug or have a question, don't hesitate to contact me
  by email at <tst@worldmaster.fr> (in french or english please)
  Thanks,
  Have fun!
]]--

--[[ VERSION=0.3.0 20110504 ]]--
local luajailversion = "luajail 0.3.0"

-- load this file only once
if luajail then return luajail end

-- module namespace in local env
local M = {}
M.verbose = false
local function PRINT(...) if M.verbose then print(...) end end

M._VERSION = luajailversion -- the module version here

assert(_VERSION and _VERSION == "Lua 5.1",
	"Version mismatch. This script are made for Lua 5.1")

assert(setfenv, "setfenv require")

---------------------------------------------------
-- source : http://www.keplerproject.org/compat/

local LUA_DIRSEP = '/'
--local LUA_OFSEP = '_'
--local OLD_LUA_OFSEP = ''
--local POF = 'luaopen_'
local LUA_PATH_MARK = '?'
--local LUA_IGMARK = ':'

local assert, error, getfenv, ipairs, loadfile, pairs, setfenv, setmetatable, type = assert, error, getfenv, ipairs, loadfile, pairs, setfenv, setmetatable, type
local loadlib = package.loadlib
local getmetatable = getmetatable

local find, format, gsub, sub = string.find, string.format, string.gsub, string.sub
local gfind = string.gmatch

local io_open = io.open

-----------------------------
--(( file system ))----------
-----------------------------

local fs = {}
fs.exists = function(file)
	PRINT("fs.exists",file)
	local f = io_open (file)
	if f then
		f:close ()
		return true
	end
	return false
end

fs.loadfile = function(file)
	PRINT("fs.loadfile", file)
	return loadfile(file)
end


-----------------------------
--(( hybrid module ))--------
-----------------------------

local _modPolicy = {}
local _modDefaultPolicy = "in"

local function modulePolicy(modname, policy)
	_modPolicy[modname] = policy
end

local function moduleDefaultPolicy(policy)
	_modDefaultPolicy = policy
end

-- for internal use
local function module_get_policy(modname)
	return _modPolicy[modname] or _modDefaultPolicy
end

local function moduleShow()
	PRINT("DefaultPolicy = "..tostring(_modDefaultPolicy))
	for k,v in pairs(_modPolicy) do
		PRINT(string.format("  %.20s = %s", k, v))
	end
end


-----------------------------
--(( jail def. ))------------
-----------------------------

local jail = {}
PRINT("jailed      _G is "..tostring(jail))
PRINT("out-of-jail _G is "..tostring(_G))
jail._G = jail
jail._VERSION = _G._VERSION -- the lua version

-- the command line arguments
jail.arg    = arg			-- ok

-- expose that is a jail ! useful for debug
jail.jailed = true
jail.jailbroken = function(obj) if obj == _G then return true end end

-----------------------------
--(( package ))--------------
-----------------------------

jail.package = {}
local _PACKAGE = jail.package

--_PACKAGE.path	= "./?.lua;./?/init.lua;"
_PACKAGE.path	= "?.lua;?/init.lua;"

_PACKAGE.cpath	= ""		-- disabled

--jail.package.cpath		= ""	-- RISK: for C lib, disable
--jail.package.loadlib = nil              -- RISK: for C lib, disable


--
-- make sure require works with standard libraries
--
jail.package.loaded = {}
local _LOADED = jail.package.loaded


--
-- avoid overwriting the package.preload table if it's already there
--
jail.package.preload = {}
local _PRELOAD = jail.package.preload


--
-- looks for a file `name' in given path
--
local function findfile (name, pname)
	name = gsub (name, "%.", LUA_DIRSEP)
	local path = _PACKAGE[pname]
	assert (type(path) == "string", format ("package.%s must be a string", pname))
	for c in gfind (path, "[^;]+") do
		c = gsub (c, "%"..LUA_PATH_MARK, name)
		if fs.exists (c) then
			return c
		end
	end
	return nil -- not found
end


--
-- check whether library is already loaded
--
--[[
local function loader_preload (name)
	assert (type(name) == "string", format (
		"bad argument #1 to `require' (string expected, got %s)", type(name)))
	assert (type(_PRELOAD) == "table", "`package.preload' must be a table")
	return _PRELOAD[name]
end
]]--

--
-- Lua library loader
--
local function loader_Lua (name)
	assert (type(name) == "string", format (
		"bad argument #1 to `require' (string expected, got %s)", type(name)))
	local filename = findfile (name, "path")
	if not filename then
		PRINT("loader_Lua not found:", filename, name)
		return false
	end
	local f, err = fs.loadfile (filename)
	if not f then
		error (format ("error loading module `%s' (%s)", name, err))
	end
	setfenv(f, jail)
	return f
end


-- findtable
local function findtable (t, f)
	assert (type(f)=="string", "not a valid field name ("..tostring(f)..")")
	local ff = f.."."
	local ok, e, w = find (ff, '(.-)%.', 1)
	while ok do
		local nt = rawget (t, w)
		if not nt then
			nt = {}
			t[w] = nt
		elseif type(t) ~= "table" then
			return sub (f, e+1)
		end
		t = nt
		ok, e, w = find (ff, '(.-)%.', e+1)
	end
	return t
end


-- package.loaders :
-- 1.  package.preload (ignore)
-- 2.  package.path    (wanted)
-- 3.  package.cpath   (required?)
-- 4.  all-in-one      (ignore)

-- create `loaders' table
--jail.package.loaders = { loader_preload, loader_Lua, loader_C, loader_Croot, }
jail.package.loaders = { loader_Lua }
-- loader_preload seems never used
-- loader_C and loader_Croot seems dangerous for the jail
local _LOADERS = jail.package.loaders


--
-- iterate over available loaders
--
local function loaders_load (name, loaders)
	-- iterate over available loaders
	assert (type (loaders) == "table", "`package.loaders' must be a table")
	for i, loader in ipairs (loaders) do
		local f = loader (name)
		if f then
			return f
		end
	end
	error (format ("module `%s' not found", name))
end

-- sentinel
local sentinel = function () end

--
-- new require
--
local function jail_require (modname)
	local _G = jail
	assert (type(modname) == "string", format (
		"bad argument #1 to `require' (string expected, got %s)", type(modname)))
	local p = _LOADED[modname]
	if p then -- is it there?
		if p == sentinel then
			error (format ("loop or previous error loading module '%s'", modname))
		end
		return p -- package is already loaded
	end
	local init = loaders_load (modname, _LOADERS)
	setfenv(init, jail)

	_LOADED[modname] = sentinel
	local actual_arg = _G.arg
	_G.arg = { modname }
	local res = init (modname)
	if res then
		_LOADED[modname] = res
	end
	_G.arg = actual_arg
	if _LOADED[modname] == sentinel then
		_LOADED[modname] = true
	end
	return _LOADED[modname]
end

local function jail_hybrid_require(name)
	--local _G = jail
	local policy = module_get_policy(name)
	PRINT(string.format("policy for module %s is '%s'", name, policy))

	if policy == "out" then
		local r = require(name)
		_LOADED[name] = package.loaded[name]
		if package.loaded[name] == _G[name] then
			jail[name] = _G[name]
		end
		return r
	elseif policy == "in" then
		return jail_require(name)
	end
	error("unknown policy for module "..name)
end


--
-- new package.seeall function
--
local function jail_package_seeall (module)
	local _G = jail
	local t = type(module)
	assert (t == "table", "bad argument #1 to package.seeall (table expected, got "..t..")")
	local meta = getmetatable (module)
	if not meta then
		meta = {}
		setmetatable (module, meta)
	end
	meta.__index = _G
end
_PACKAGE.seeall = jail_package_seeall

--
-- new module function
--
local function jail_module (modname, ...)
	local _G = jail
	local ns = _LOADED[modname]
	if type(ns) ~= "table" then
		ns = findtable (_G, modname)
		if not ns then
			error (string.format ("name conflict for module '%s'", modname))
		end
		_LOADED[modname] = ns
	end
	if not ns._NAME then
		ns._NAME = modname
		ns._M = ns
		ns._PACKAGE = gsub (modname, "[^.]*$", "")
	end
	setfenv (2, ns)
	for i, f in ipairs (arg) do
		f (ns)
	end
end


---------------------------------------------------

-- the problem explained here :
--   http://stackoverflow.com/questions/3098544/lua-variable-scoping-with-setfenv
-- require behavior : http://www.lua.org/pil/8.1.html

local function jail_dofile(name)
	local f,e = fs.loadfile(name)
	if not f then error(e, 2) end
PRINT("dofile env:", jail, getfenv(2))
	--setfenv(f, getfenv(2))
	setfenv(f, jail)
	return f()
end

local function jail_loadfile(name)
	local f,e = fs.loadfile(name)
	if not f then error(e, 2) end
	--setfenv(f, getfenv(2))
PRINT("jail_loadfile env:", jail, getfenv(2))
	setfenv(f, jail)
	return f
end

--local function jail_dofile(name) return jail_loadfile(name)() end



local function jail_print(...) print("jail:", ...) end

local function Hprint(...) print(" host:", ...) end

--local function pack(...) return {...} end



-----------------------------
--[[ lua functions ]]--------
-----------------------------



-- getfenv() or getfenv(0) return the out-of-jail _G
local function jail_getfenv(f)
	if type(f) == "function" then
		return getfenv(f)
	end
	-- else error() ?
	return jail -- jail._G
end

local function jail_subst(...)
	local a = {...}
	for i,o in ipairs(a) do
		if o == _G then
			a[i]=jail
		end
	end
	return unpack(a)
end
--print(">>", _G, nil, jail, "a")
--print(">>", jail_subst(_G, nil, jail, "a"))

local function jail_getmetatable(obj)
	return jail_subst(getmetatable(jail_subst(obj)))
end

local function jail_setmetatable(t, mt)
	return jail_subst(setmetatable(jail_subst(t, mt)))
end


jail.assert		= assert		-- ok
jail.collectgarbage	= collectgarbage	-- seems ok
jail.dofile		= jail_dofile		-- RISK: catched
jail.error		= error			-- seems ok
jail.getfenv		= jail_getfenv		-- RISK: lua env jailbreak risk
jail.getmetatable	= jail_getmetatable		-- ok
jail.ipairs		= ipairs		-- ok
jail.load		= load			-- seems ok
jail.loadfile		= jail_loadfile		-- RISK: catched
jail.loadstring		= loadstring		-- seems ok
jail.module		= jail_module		-- RISK: catched ?
jail.next		= next			-- ok
jail.pairs		= pairs			-- ok
jail.pcall		= pcall			-- ok
jail.print		= jail_print		-- ok (for debug)
jail.rawequal		= rawequal		-- ok
jail.rawget		= rawget		-- ok
jail.rawset		= rawset		-- ok
jail.require		= jail_hybrid_require		-- RISK: catched ?
jail.select		= select		-- ok
--setfenv		= setfenv		-- seems not ok, is able to reset the current env ?
jail.setmetatable	= setmetatable		-- ok
jail.tonumber		= tonumber		-- ok
jail.tostring		= tostring		-- ok
jail.type		= type			-- ok
jail.unpack		= unpack		-- ok
jail.xpcall		= xpcall		-- ok


-----------------------------
--(( os ))-------------------
-----------------------------

jail.os = {}
jail.os.clock     = os.clock		-- ok
jail.os.date      = os.date		-- ok
jail.os.difftime  = os.difftime		-- ok
jail.os.getenv    = os.getenv		-- ok
jail.os.setlocale = os.setlocale	-- seems ok
jail.os.time      = os.time		-- ok
--jail.os.execute   = os.execute	-- RISK
--jail.os.exit      = os.exit		-- ?
--jail.os.remove    = os.remove		-- RISK
--jail.os.rename    = os.rename		-- RISK
--jail.os.tmpname   = os.tmpname	-- RISK


-----------------------------
--(( io ))-------------------
-----------------------------

jail.io = {}
jail.io.flush        = io.flush		-- seems ok
-- RISK: file loading jailbreak risk	-- patched, ok
jail.io.input = function() -- no file argument
        return io.input()
end

-- RISK: file writing jailbreak risk	-- patched, ok
jail.io.output       = function() -- no file argument
        return io.output()
end
jail.io.read         = io.read -- really ?
jail.io.stderr       = io.stderr	-- ok
jail.io.stdin        = io.stdin		-- ok
jail.io.stdout       = io.stdout	-- ok
jail.io.type         = io.type		-- ok
jail.io.write        = io.write		-- seems ok (is for stdout, not file)
jail.io.open         = io.open		-- ok
jail.io.close        = io.close		-- ok
jail.io.lines        = io.lines		-- ok
--jail.io.popen        = io.popen	-- RISK
jail.io.tmpfile      = io.tmpfile	-- ok


-----------------------------
--(( math ))-----------------
-----------------------------

jail.math   = math			-- ok


-----------------------------
--(( string ))---------------
-----------------------------

jail.string = string			-- ok


-----------------------------
--(( table ))----------------
-----------------------------

jail.table  = table			-- ok


-----------------------------
--(( coroutine ))------------
-----------------------------

--coroutine				-- seems not ok
jail.coroutine = coroutine		--TOCHECK


-----------------------------
--(( debug ))----------------
-----------------------------

--debug					-- seems not ok
jail.debug = {}
jail.debug.traceback = debug.traceback	-- TOCHECK
jail.debug.getinfo = debug.getinfo	-- TOCHECK


-----------------------------
--(( package.loaded ))-------
-----------------------------

_LOADED._G        = jail._G
_LOADED.package   = jail.package
_LOADED.debug     = jail.debug
_LOADED.string    = jail.string
_LOADED.math      = jail.math
_LOADED.io        = jail.io
_LOADED.os        = jail.os
_LOADED.table     = jail.table 
_LOADED.coroutine = jail.coroutine


-----------------------------
---- jail facilities --------
-----------------------------

local function registermodule(modname, modobj)
	assert(type(modname) == "string", "modname must be a string")
	assert(modobj, "modobj must not be nil")
	--if not jail then return end
	if jail.package and jail.package.loaded then
		if jail.package.loaded[modname] then
			error ("module already loaded : "..modname)
		end
		jail.package.loaded[modname] = modobj
	end
	if jail[modname] then
		error ("module already loaded : "..modname)
	end
	jail[modname] = modobj
end

local function unregistermodule(modname)
	if jail and jail.package and jail.package.loaded and jail.package.loaded[modname] then
		jail.package.loaded[modname] = nil
		if jail[modname] then
			jail[modname] = nil
		end
	end
end


------------------------------------------------

M._VERSION = luajailversion -- the module version here
M.jail = jail -- allow to modify the jail env.
M.fs   = fs   -- allow to customise the file system handler

M.modulePolicy		= modulePolicy
M.moduleDefaultPolicy	= moduleDefaultPolicy
M.moduleShow		= moduleShow

M.disable = false
M.unregistermodule = unregistermodule
M.registermodule = registermodule

M.setup = function(f)
	if not M.disable then
		setfenv(f, jail)
	end
	return f
end

M.run = function(f, ...)
	if not M.disable then
		setfenv(f, jail)
	end
	return f(...)
end

-- setup the module namespace in global env
luajail = M
return M -- also return the module namespace
